/*
  (c) 2003-2020 Philipp Klaus Krause philipp@colecovision.eu, pkk@spth.de.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.
   
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
   
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#include <string>
#include <iostream>
#include <fstream>
#include <utility>
#include <math.h>
#include <sys/types.h>
#include <stdint.h>
#include <string.h>

#include <wand/MagickWand.h>

// Default palette: Colecovision RGBA palette.
double palette[16][4] = {
	{0 / 7.0, 0 / 7.0, 0 / 7.0, 0 / 7.0},
	{0 / 7.0, 0 / 7.0, 0 / 7.0, 7 / 7.0},
	{1 / 7.0, 6 / 7.0, 1 / 7.0, 7 / 7.0},
	{3 / 7.0, 7 / 7.0, 3 / 7.0, 7 / 7.0},
	{1 / 7.0, 1 / 7.0, 7 / 7.0, 7 / 7.0},
	{2 / 7.0, 3 / 7.0, 7 / 7.0, 7 / 7.0},
	{5 / 7.0, 1 / 7.0, 1 / 7.0, 7 / 7.0},
	{2 / 7.0, 6 / 7.0, 7 / 7.0, 7 / 7.0},
	{7 / 7.0, 1 / 7.0, 1 / 7.0, 7 / 7.0},
	{7 / 7.0, 3 / 7.0, 3 / 7.0, 7 / 7.0},
	{6 / 7.0, 6 / 7.0, 1 / 7.0, 7 / 7.0},
	{6 / 7.0, 6 / 7.0, 4 / 7.0, 7 / 7.0},
	{1 / 7.0, 4 / 7.0, 1 / 7.0, 7 / 7.0},
	{6 / 7.0, 2 / 7.0, 5 / 7.0, 7 / 7.0},
	{5 / 7.0, 5 / 7.0, 5 / 7.0, 7 / 7.0},
	{7 / 7.0, 7 / 7.0, 7 / 7.0, 7 / 7.0}};

class output
{
public:
	output(std::ostream& o);
	~output(void);
	void set_binary(bool b);
	void set_name(const std::string& n);
	void write(uint8_t b);
private:
	std::ostream *out;
	bool binary;
	std::string name;
	bool started;
};

output::output(std::ostream& o)
{
	out = &o;
	binary = false;
	name = "data";
	started = false;
}

output::~output(void)
{
	if(!binary)
		(*out) << "};\n";
}

void output::set_binary(bool b)
{
	binary = b;
}

void output::set_name(const std::string& n)
{
	name = n;
}

void output::write(uint8_t b)
{
	if(!started)
	{
		if(!binary)
		{
			(*out) << "const unsigned char " << name << "[] = {";
			out->setf(std::ios_base::hex, std::ios_base::basefield);
			out->setf(std::ios_base::showbase);
		}
		
	}

	if(!binary)
	{	
		if(started)
			(*out) << ", ";
		(*out) << int(b);
	}
	else
		out->put(b);

	started = true;
}

// Convert 8 RGB pixels (one row of character).
inline void rgba2sms(PixelWand **in, u_int8_t *pattern)
{
	double row[8][4];
	unsigned int matches[8];

	for(unsigned int i = 0; i < 8; i++)  // Conversion to 9 bit color.
	{
		row[i][0] = PixelGetRed(in[i]);
		row[i][1] = PixelGetGreen(in[i]);
		row[i][2] = PixelGetBlue(in[i]);
		row[i][3] = PixelGetAlpha(in[i]);
	}

	// Find best-matchig color in palette for each pixel in row.
	for(unsigned int k = 0; k < 8; k++)
	{
		double dists[16];

		for(unsigned int i = 0; i < 16; i++)
		{
			dists[i] = 0;
			dists[i] += fabs(palette[i][0] - row[k][0]);
			dists[i] += fabs(palette[i][1] - row[k][1]);
			dists[i] += fabs(palette[i][2] - row[k][2]);
			dists[i] += fabs(palette[i][3] - row[k][3]) * 8.0;
		}

		unsigned int i;
		for(i = 1, matches[k] = 0; i < 16; i++)
			if (dists[i] < dists[matches[k]])
				matches[k] = i;
	}

	memset(pattern, 0, 4);
	for(unsigned int i = 0; i < 4; i++)
		for(unsigned int k = 0; k < 8; k++)
			if(matches[k] & (1 << i))
				pattern[i] |= (1 << ( 7 - k));
}

// Convert 8 RGB pixels (one row of character).
inline void rgb2sms(PixelWand **in, u_int8_t *pattern)
{
	double row[8][3];
	unsigned int matches[8];

	for(unsigned int i = 0; i < 8; i++)  // Read row of 8 pixels
	{
		row[i][0] = PixelGetRed(in[i]);
		row[i][1] = PixelGetGreen(in[i]);
		row[i][2] = PixelGetBlue(in[i]);
	}

	// Find best-matchig color in palette for each pixel in row.
	for(unsigned int k = 0; k < 8; k++)
	{
		double dists[16];

		for(unsigned int i = 1; i < 16; i++)
		{
			dists[i] = 0;
			dists[i] += fabs(palette[i][0] - row[k][0]);
			dists[i] += fabs(palette[i][1] - row[k][1]);
			dists[i] += fabs(palette[i][2] - row[k][2]);
		}

		unsigned int i;
		for(i = 2, matches[k] = 1; i < 16; i++)
			if (dists[i] < dists[matches[k]])
				matches[k] = i;
	}

	memset(pattern, 0, 4);
	for(unsigned int i = 0; i < 4; i++)
		for(unsigned int k = 0; k < 8; k++)
			if(matches[k] & (1 << i))
				pattern[i] |= (1 << ( 7 - k));
}

// Convert 8 RGBA pixels (one row of character).
inline void rgba2cv(PixelWand **in, u_int8_t *pattern, u_int8_t *color)
{
	float row[8][4];
	for(unsigned int i = 0; i < 8; i++)  // Read row of 8 pixels.
	{
		row[i][0] = PixelGetRed(in[i]);
		row[i][1] = PixelGetGreen(in[i]);
		row[i][2] = PixelGetBlue(in[i]);
		row[i][3] = PixelGetAlpha(in[i]);
	}

	double dist[16][16];

	// Find best two colors
	// Calculate differences
	for(unsigned int i = 0; i < 16; i++)
	{
		for(unsigned int j = 0; j < 16; j++)
		{
			dist[i][j] = 0;
			for(unsigned int k = 0; k < 8; k++)
			{
				if((palette[i][3] != 0) && (palette[j][3] != 0) || (row[k][3] != 0))
				{
					dist[i][j] += std::min(fabs(palette[i][0] - row[k][0]), fabs(palette[j][0] - row[k][0]));
					dist[i][j] += std::min(fabs(palette[i][1] - row[k][1]), fabs(palette[j][1] - row[k][1]));
					dist[i][j] += std::min(fabs(palette[i][2] - row[k][2]), fabs(palette[j][2] - row[k][2]));
				}
				dist[i][j] += std::min(fabs(palette[i][3] - row[k][3]), fabs(palette[j][3] - row[k][3]));
			}
		}
	}

  for(float fd = 0.0f;; fd+= 0.1f)
  {
    for(unsigned int j = 0; j < 16; j++)
      for(unsigned int l = 0; l < 16; l++)
        if(dist[j][l] <= fd)  // Best match
        {
          *color = ((j << 4) + l);
          *pattern = 0;
          for(unsigned int k = 0; k < 8; k++)
          {
            (*pattern) <<= 1;
            double dist1 = 0, dist2 = 0;
            dist1 += fabs(palette[j][0] - row[k][0]);
            dist2 += fabs(palette[l][0] - row[k][0]);
            dist1 += fabs(palette[j][1] - row[k][1]);
            dist2 += fabs(palette[l][1] - row[k][1]);
            dist1 += fabs(palette[j][2] - row[k][2]);
            dist2 += fabs(palette[l][2] - row[k][2]);
			dist1 += fabs(palette[j][2] - row[k][2]) * 8.0;
            dist2 += fabs(palette[l][2] - row[k][2]) * 8.0;
            if(dist1 <= dist2) (*pattern)++;     
          }
          return;
        }
  }
  return;
}

// Convert 8 RGB pixels (one row of character).
inline void rgb2cv(PixelWand **in, u_int8_t *pattern, u_int8_t *color)
{
	static unsigned int oj = 1, ol = 1;	// Try to reuse old color combination to optimize output for compression.
	static u_int8_t opattern;	// Try to reuse old pattern to optimize output for compression.

	unsigned int j, l;
	double row[8][3];
    
	for(unsigned int i = 0; i < 8; i++)  // Read row of 8 pixels
	{
		row[i][0] = PixelGetRed(in[i]);
		row[i][1] = PixelGetGreen(in[i]);
		row[i][2] = PixelGetBlue(in[i]);
	}

	double dist[16][16];

	// Find best two colors
	// Calculate differences
	for(unsigned int i = 1; i < 16; i++)
		for(j = 1; j < 16; j++)
		{
			dist[i][j] = 0;
			for(unsigned int k = 0; k < 8; k++)
			{
				double d0 = 0.0, d1 = 0.0;
				d0 += fabs(palette[i][0] - row[k][0]);
				d1 += fabs(palette[j][0] - row[k][0]);
				d0 += fabs(palette[i][1] - row[k][1]) * 1.5;
				d1 += fabs(palette[j][1] - row[k][1]) * 1.5;
				d0 += fabs(palette[i][2] - row[k][2]);
				d1 += fabs(palette[j][2] - row[k][2]);
				dist[i][j] += std::min(d0, d1);
			}
		}

	for(float fd = 0;;fd += 0.025f)
	{
		if(dist[oj][ol] <= fd)  // Try old combination first to optimize output for compression.
		{
			j = oj;
			l = ol;
			goto done;
		}

		for(j = 1; j < 16; j++)
			for(l = 1; l < 16; l++)
				if(dist[j][l] <= fd)  // Best match
					goto done;
	}
done:
	*pattern = 0;

	for(unsigned int k = 0; k < 8; k++)
	{
		(*pattern) <<= 1;
		double dist1 = 0.0, dist2 = 0.0;
		dist1 += fabs(palette[j][0] - row[k][0]);
		dist2 += fabs(palette[l][0] - row[k][0]);
		dist1 += fabs(palette[j][1] - row[k][1]);
		dist2 += fabs(palette[l][1] - row[k][1]);
		dist1 += fabs(palette[j][2] - row[k][2]);
		dist2 += fabs(palette[l][2] - row[k][2]);
		if(dist1 <= dist2) (*pattern)++;
	}

	if(*pattern == (uint8_t)(~opattern) && (oj != j || ol != l))
	{
		int tmp;
		*pattern = opattern;
		tmp = j;
		j = l;
		l = tmp;
	}
	*color = (j << 4) + l;
	oj = j;
	ol = l;
	opattern = *pattern;

	return;
}

inline void rgba2bw(const void *in, u_int8_t *pattern)
{
}

inline void rgb2bw(const void *in, u_int8_t *pattern)
{
}

void load_palette(const char *palettefile)
{
	MagickWand *wand = NewMagickWand();  
  
	if(!MagickReadImage(wand, palettefile))
	{
		std::cerr << "Warning: Failed to open palette image " << palettefile << ". Custom palette ignored.\n";
		return;
	}

	size_t width = MagickGetImageWidth(wand);
	size_t height = MagickGetImageHeight(wand);

	if(width * height != 16)
	{
		std::cerr << "Warning: Palette file needs to have exactly 16 pixels. Custom palette ignored.\n";
		return;		
	}
	
	PixelIterator *p = NewPixelIterator(wand);

	int i = 0;
	for(size_t j = 0; j < height; j++)
	{
		PixelWand **pixels = PixelGetNextIteratorRow(p, &width);

		for(size_t k = 0; k < width; k++)
		{
			palette[i][0] = PixelGetRed(pixels[k]);
			palette[i][1] = PixelGetGreen(pixels[k]);
			palette[i][2] = PixelGetBlue(pixels[k]);
			palette[i][3] = PixelGetAlpha(pixels[k]);
		}
	}

	DestroyPixelIterator(p);
	
	DestroyMagickWand(wand);
}

// Convert color channel to 2 bit precision
static unsigned char c22(double color)
{
	unsigned char ret = color * 3.0 + 0.5;
	return (ret);
}

void export_palette(const char *palettefile, bool binary)
{
	std::cerr << "Pallete export not implementd.\n";

	std::ofstream paletteo;

	paletteo.open(palettefile, std::ios::out | std::ios::trunc);
	
	output *paletteout;
	{
		paletteout = new output(paletteo);
		paletteout->set_name("palette");
		paletteout->set_binary(binary);
	}

	for(int i = 0; i < 16; i++)
	{
		uint8_t entry = c22(palette[i][0]) + (c22(palette[i][1]) << 2) + (c22(palette[i][2]) << 4);
		paletteout->write(entry);
	}

	delete paletteout;
	
	paletteo.close();
}

void usage(void)
{
	std::cerr
		<< '\n' << PACKAGE << ' ' << PACKAGE_VERSION << " (c) 2003-2016 Philipp Klaus Krause " << PACKAGE_BUGREPORT
		<< "\nUsage: png2cv [-b][--binary][--monochrome][--verbose] file\n"
		<< "\n-b, --binary: binary output instead of C source."
		<< "\n--bitmap: Generate data for bitmap mode (default)."
		<< "\n--bitmap-text: Generate data for bitmap text mode."
		<< "\n--4: Generate data for mode 4 (SMS)."
		<< "\n--palette <file>: Use file as palette."
		<< "\n--export-palette4 <file>: Export palette data for mode 4 (SMS)."
		<< "\n--verbose: Verbose output.\n";
	std::cerr << "The --bitmap-text option is currently broken!\n";
}

int main(int argc, char** argv)
{
	std::string ifile, patternofile, colorofile;
	bool binary = false;
	bool color = true;
	bool verbose = false;
	bool mode4 = false;

	int i;

	if(argc == 1)
	{
		std::cerr << "Too few arguments.\n";
		usage();
		return(-1);
	}

	MagickWandGenesis();

	for(i = 1; i < argc - 1; i++)
	{
		std::string stra = argv[i];
		if(stra == "-b")
			binary = true;
		else if(stra == "--binary")
			binary = true;
		else if(stra == "--bitmap")
			color = true;
		else if(stra == "--bitmap-text")
			color = false;
		else if(stra == "--4")
			mode4 = true;
		else if(stra == "--palette" && i < argc - 2) 
			load_palette(argv[++i]);
		else if(stra == "--export-palette4" && i < argc - 2)
			export_palette(argv[++i], binary);
		else if(stra == "--verbose")
			verbose = true;
		else
		{
			usage();
			MagickWandTerminus();
			return(-1);
		}
	}

	ifile = argv[argc - 1];

	patternofile = ifile;
	colorofile = ifile;
	if(binary)
	{
		patternofile += mode4 ? ".pattern4.bin" : ".pattern.bin";
		colorofile += ".color.bin"; 
	}
	else
	{
		patternofile += mode4 ? ".pattern4.c" : ".pattern.c";
		colorofile += ".color.c";
	}
  
	MagickWand *wand = NewMagickWand();  
  
	if(!MagickReadImage(wand, ifile.c_str()))
	{
		std::cerr << "Failed to open input image " << ifile << "\n";
		return(-1);
	}

	size_t width = MagickGetImageWidth(wand);
	size_t height = MagickGetImageHeight(wand);
	bool alpha = MagickGetImageAlphaChannel(wand);

	if((width % 8) || (height % 8))
	{
		std::cerr << "Width and height must be multiples of 8.";
		return(-1);
	}

	u_int8_t cvimage[4];
	u_int8_t cvcolor;
	std::ofstream patterno;
	std::ofstream coloro;

	patterno.open(patternofile.c_str(), std::ios::out | std::ios::trunc);
	if(color && !mode4) coloro.open(colorofile.c_str(), std::ios::out | std::ios::trunc);

	output *patternout;
	output *colorout;
	{
		patternout = new output(patterno);
		patternout->set_name("pattern");
		patternout->set_binary(binary);
	}
	if(color && !mode4)
	{
		colorout = new output(coloro);
		colorout->set_name("color");
		colorout->set_binary(binary);
	}

	if(verbose)
	{
		std::cout << "Converting " << width << "x" << height << " " << (alpha ? "RGBA" : "RGB") << " image.\n";
		std::cout << "A . for every Colecovision character line written.\n";
		std::cout.flush();
	}

	PixelIterator *p = NewPixelIterator(wand);

	for(size_t i = 0; i < height; i += 8)
		for(size_t j = 0; j < width; j += 8)
		{
			for(size_t k = 0; k < 8; k++)
			{
				PixelSetIteratorRow(p, i + k);
				size_t num_pixel_wands;

				if(mode4)
				{
					if(alpha) // RGBA
						rgba2sms(PixelGetCurrentIteratorRow(p, &num_pixel_wands) + j, cvimage);
					else  // RGB
						rgb2sms(PixelGetCurrentIteratorRow(p, &num_pixel_wands) + j, cvimage);
				}
				else if(color)
				{
					if(alpha) // RGBA
						rgba2cv(PixelGetCurrentIteratorRow(p, &num_pixel_wands) + j, cvimage, &cvcolor);
					else  // RGB
						rgb2cv(PixelGetCurrentIteratorRow(p, &num_pixel_wands) + j, cvimage, &cvcolor);
				}
				else
 				{
					if(alpha) // RGBA
						rgba2bw(PixelGetCurrentIteratorRow(p, &num_pixel_wands) + j, cvimage);
 					else  // RGB
						rgb2bw(PixelGetCurrentIteratorRow(p, &num_pixel_wands) + j, cvimage);
				}

				if(!mode4)
					patternout->write(*cvimage);
				else
					for(int i = 0; i < 4; i++)
						patternout->write(cvimage[i]);

				if(color && !mode4)
					colorout->write(cvcolor);
			}
			if(verbose)
			{
				std::cout << '.';
				std::cout.flush();
			}
		}
		
	DestroyPixelIterator(p);
	
	DestroyMagickWand(wand);

	delete patternout;
	if(color && !mode4)
		delete colorout;
	patterno.close();
	if(color && !mode4)
		coloro.close();
	if(verbose) std::cout << "\ndone.\n";

	MagickWandTerminus();

	return(0);
}

